import grpc
from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import DataApp as MDDataApp
from github.com.metaprov.modelaapi.services.dataapp.v1.dataapp_pb2_grpc import DataAppServiceStub
from github.com.metaprov.modelaapi.services.dataapp.v1.dataapp_pb2 import CreateDataAppRequest, \
    UpdateDataAppRequest, \
    DeleteDataAppRequest, GetDataAppRequest, ListDataAppsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class DataApp(Resource):
    def __init__(self, item: MDDataApp = MDDataApp(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class DataAppClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DataAppServiceStub = stub

    def create(self, dataapp: DataApp) -> bool:
        request = CreateDataAppRequest()
        request.dataapp.CopyFrom(dataapp.raw_message)
        try:
            response = self.__stub.CreateDataApp(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, dataapp: DataApp) -> bool:
        request = UpdateDataAppRequest()
        request.dataapp.CopyFrom(dataapp.raw_message)
        try:
            self.__stub.UpdateDataApp(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[DataApp, bool]:
        request = GetDataAppRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataApp(request)
            return DataApp(response.dataapp, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDataAppRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataApp(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[DataApp], bool]:
        request = ListDataAppsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDataApps(request)
            return [DataApp(item, self) for item in response.dataapps.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import VirtualCluster as MDVirtualCluster
from github.com.metaprov.modelaapi.services.virtualcluster.v1.virtualcluster_pb2_grpc import VirtualClusterServiceStub
from github.com.metaprov.modelaapi.services.virtualcluster.v1.virtualcluster_pb2 import CreateVirtualClusterRequest, \
    UpdateVirtualClusterRequest, \
    DeleteVirtualClusterRequest, GetVirtualClusterRequest, ListVirtualClustersRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class VirtualCluster(Resource):
    def __init__(self, item: MDVirtualCluster = MDVirtualCluster(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class VirtualClusterClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: VirtualClusterServiceStub = stub

    def create(self, virtualcluster: VirtualCluster) -> bool:
        request = CreateVirtualClusterRequest()
        request.virtualcluster.CopyFrom(virtualcluster.raw_message)
        try:
            response = self.__stub.CreateVirtualCluster(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, virtualcluster: VirtualCluster) -> bool:
        request = UpdateVirtualClusterRequest()
        request.virtualcluster.CopyFrom(virtualcluster.raw_message)
        try:
            self.__stub.UpdateVirtualCluster(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[VirtualCluster, bool]:
        request = GetVirtualClusterRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetVirtualCluster(request)
            return VirtualCluster(response.virtualcluster, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteVirtualClusterRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteVirtualCluster(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[VirtualCluster], bool]:
        request = ListVirtualClustersRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListVirtualClusters(request)
            return [VirtualCluster(item, self) for item in response.virtualclusters.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





from dataclasses import dataclass
from enum import Enum

from github.com.metaprov.modelaapi.services.common.v1.common_pb2 import Plot
from k8s.io.api.core.v1.generated_pb2 import ObjectReference, SecretReference
from k8s.io.apimachinery.pkg.apis.meta.v1.generated_pb2 import Time

from modela.Configuration import Configuration, ImmutableConfiguration, datamodel


class ConditionStatus(Enum):
    ConditionTrue = "True"
    ConditionFalse = "False"
    ConditionUnknown = "Unknown"


class TaskName(Enum):
    BatchPredictTask = "batch-predict"
    ForecastTask = "forecast"
    ProfileDatasetTask = "profile-dataset"
    SnapshotDatasetTask = "snapshot-dataset"
    ProfileStudyTask = "profile-study"
    ProfileModelTask = "profile-model"
    PublishModelTask = "publish-model"
    BakeModelTask = "bake-model"
    RunRecipeTask = "run-recipe"
    ReportDatasetTask = "report-dataset"
    ReportModelTask = "report-model"
    ReportStudyTask = "report-study"
    ReportPeriodTask = "report-period"
    ReportSummaryTask = "report-summary"
    SplitDatasetTask = "split-dataset"
    TestEnsembleTask = "test-ensemble"
    TestModelTask = "test-model"
    TrainEnsembleTask = "train-ensemble"
    TrainModelTask = "train-model"
    ValidateDatasetTask = "validate-dataset"
    MergeForecastTask = "merge-forecast"
    PartitionForecastTask = "partition-forecast"
    SplitDatasetToRungs = "split-dataset-to-rungs"
    CompileTask = "compile-model"
    PackageModelTask = "package-model"
    FeatureGenTask = "feature-gen"
    FeatureSelectTask = "feature-select"
    RunSqlQueryTask = "run-sql-query"
    RunWebRequestTask = "run-web-request"
    GenerateDataset = "generate-dataset"
    ValidateModel = "validate-model"


class StatusError(Enum):
    InvalidConfiguration = "InvalidConfiguration"
    InsufficientResources = "InsufficientResources"
    CreateError = "CreateError"
    UpdateError = "UpdateError"
    DeleteError = "DeleteError"


class PriorityLevel(Enum):
    Low = "low"
    Medium = "medium"
    High = "high"
    Urgent = "urgent"


class HardwareTarget(Enum):
    AMDEPYC2 = "amd-epyc-2"
    ARMA53 = "arma-53"
    ARMA72 = "arma-72"
    IntelCascadeLake = "intel-cascade-lake"
    IntelBroadwell = "intel-broadwell"
    IntelSkylake = "intel-skylake"
    TeslaV100 = "tesla-v100"
    TeslaK80 = "tesla-k80"
    T4 = "t4"
    RaspberryPi8MODELB = "raspberry-pi8-modela-b"


class TriggerScheduleEventType(Enum):
    Now = "now"
    Once = "once"
    Hourly = "hourly"
    Daily = "daily"
    Weekly = "weekly"
    Monthly = "monthly"
    Yearly = "yearly"
    Cron = "cron"


class Freq(Enum):
    Seconds = "second"
    Minutes = "minute"
    Hours = "hour"
    Days = "day"
    Weeks = "week"
    Months = "month"
    Qtrs = "quarter"
    Years = "year"


@datamodel(proto=Time)
class Time(Configuration):
    Seconds: int = 0
    Nanos: int = 0


@datamodel(proto=ObjectReference)
class ObjectReference(Configuration):
    """
    An Object Reference defines the location of a resource on the current operational cluster, denoted by the namespace
    of the resource and the name of the resource which exists on that namespace.
    """

    Namespace: str = ""
    Name: str = ""


@datamodel(proto=SecretReference)
class SecretReference(Configuration):
    Name: str = ""
    Namespace: str = ""

@datamodel(proto=Plot)
class Plot(ImmutableConfiguration):
    Fname: str = ""
    Img: bytes = ""
    Name: str = ""
    Title: str = ""
    Url: str = ""




import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import LabelingPipelineRun as MDLabelingPipelineRun
from github.com.metaprov.modelaapi.services.labelingpipelinerun.v1.labelingpipelinerun_pb2_grpc import LabelingPipelineRunServiceStub
from github.com.metaprov.modelaapi.services.labelingpipelinerun.v1.labelingpipelinerun_pb2 import CreateLabelingPipelineRunRequest, \
    UpdateLabelingPipelineRunRequest, \
    DeleteLabelingPipelineRunRequest, GetLabelingPipelineRunRequest, ListLabelingPipelineRunRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class LabelingPipelineRun(Resource):
    def __init__(self, item: MDLabelingPipelineRun = MDLabelingPipelineRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class LabelingPipelineRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: LabelingPipelineRunServiceStub = stub

    def create(self, labelingpipelinerun: LabelingPipelineRun) -> bool:
        request = CreateLabelingPipelineRunRequest()
        request.labelingpipelinerun.CopyFrom(labelingpipelinerun.raw_message)
        try:
            response = self.__stub.CreateLabelingPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, labelingpipelinerun: LabelingPipelineRun) -> bool:
        request = UpdateLabelingPipelineRunRequest()
        request.labelingpipelinerun.CopyFrom(labelingpipelinerun.raw_message)
        try:
            self.__stub.UpdateLabelingPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[LabelingPipelineRun, bool]:
        request = GetLabelingPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetLabelingPipelineRun(request)
            return LabelingPipelineRun(response.labelingpipelinerun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteLabelingPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteLabelingPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[LabelingPipelineRun], bool]:
        request = ListLabelingPipelineRunRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListLabelingPipelineRuns(request)
            return [LabelingPipelineRun(item, self) for item in response.labelingpipelineruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import SqlQueryRun as MDSqlQueryRun
from github.com.metaprov.modelaapi.services.sqlqueryrun.v1.sqlqueryrun_pb2_grpc import SqlQueryRunServiceStub
from github.com.metaprov.modelaapi.services.sqlqueryrun.v1.sqlqueryrun_pb2 import CreateSqlQueryRunRequest, \
    UpdateSqlQueryRunRequest, \
    DeleteSqlQueryRunRequest, GetSqlQueryRunRequest, ListSqlQueryRunsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class SqlQueryRun(Resource):
    def __init__(self, item: MDSqlQueryRun = MDSqlQueryRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class SqlQueryRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: SqlQueryRunServiceStub = stub

    def create(self, sqlqueryrun: SqlQueryRun) -> bool:
        request = CreateSqlQueryRunRequest()
        request.sqlqueryrun.CopyFrom(sqlqueryrun.raw_message)
        try:
            response = self.__stub.CreateSqlQueryRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, sqlqueryrun: SqlQueryRun) -> bool:
        request = UpdateSqlQueryRunRequest()
        request.sqlqueryrun.CopyFrom(sqlqueryrun.raw_message)
        try:
            self.__stub.UpdateSqlQueryRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[SqlQueryRun, bool]:
        request = GetSqlQueryRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetSqlQueryRun(request)
            return SqlQueryRun(response.sqlqueryrun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteSqlQueryRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteSqlQueryRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[SqlQueryRun], bool]:
        request = ListSqlQueryRunsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListSqlQueryRuns(request)
            return [SqlQueryRun(item, self) for item in response.sqlqueryruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import VirtualBucket as MDVirtualBucket
from github.com.metaprov.modelaapi.services.virtualbucket.v1.virtualbucket_pb2_grpc import VirtualBucketServiceStub
from github.com.metaprov.modelaapi.services.virtualbucket.v1.virtualbucket_pb2 import CreateVirtualBucketRequest, \
    UpdateVirtualBucketRequest, \
    DeleteVirtualBucketRequest, GetVirtualBucketRequest, ListVirtualBucketsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class VirtualBucket(Resource):
    def __init__(self, item: MDVirtualBucket = MDVirtualBucket(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class VirtualBucketClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: VirtualBucketServiceStub = stub

    def create(self, virtualbucket: VirtualBucket) -> bool:
        request = CreateVirtualBucketRequest()
        request.virtualbucket.CopyFrom(virtualbucket.raw_message)
        try:
            response = self.__stub.CreateVirtualBucket(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, virtualbucket: VirtualBucket) -> bool:
        request = UpdateVirtualBucketRequest()
        request.virtualbucket.CopyFrom(virtualbucket.raw_message)
        try:
            self.__stub.UpdateVirtualBucket(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[VirtualBucket, bool]:
        request = GetVirtualBucketRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetVirtualBucket(request)
            return VirtualBucket(response.virtualbucket, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteVirtualBucketRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteVirtualBucket(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[VirtualBucket], bool]:
        request = ListVirtualBucketsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListVirtualBuckets(request)
            return [VirtualBucket(item, self) for item in response.virtualbuckets.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import License as MDLicense
from github.com.metaprov.modelaapi.services.license.v1.license_pb2_grpc import LicenseServiceStub
from github.com.metaprov.modelaapi.services.license.v1.license_pb2 import CreateLicenseRequest, \
    UpdateLicenseRequest, \
    DeleteLicenseRequest, GetLicenseRequest, ListLicensesRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class License(Resource):
    def __init__(self, item: MDLicense = MDLicense(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class LicenseClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: LicenseServiceStub = stub

    def create(self, license: License) -> bool:
        request = CreateLicenseRequest()
        request.license.CopyFrom(license.raw_message)
        try:
            response = self.__stub.CreateLicense(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, license: License) -> bool:
        request = UpdateLicenseRequest()
        request.license.CopyFrom(license.raw_message)
        try:
            self.__stub.UpdateLicense(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[License, bool]:
        request = GetLicenseRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetLicense(request)
            return License(response.license, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteLicenseRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteLicense(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[License], bool]:
        request = ListLicensesRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListLicenses(request)
            return [License(item, self) for item in response.licenses.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Featureset as MDFeatureset
from github.com.metaprov.modelaapi.services.featureset.v1.featureset_pb2_grpc import FeaturesetServiceStub
from github.com.metaprov.modelaapi.services.featureset.v1.featureset_pb2 import CreateFeaturesetRequest, \
    UpdateFeaturesetRequest, \
    DeleteFeaturesetRequest, GetFeaturesetRequest, ListFeaturesetRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Featureset(Resource):
    def __init__(self, item: MDFeatureset = MDFeatureset(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name)


class FeaturesetClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: FeaturesetServiceStub = stub

    def create(self, featureset: Featureset) -> bool:
        request = CreateFeaturesetRequest()
        request.featureset.CopyFrom(featureset.raw_message)
        try:
            response = self.__stub.CreateFeatureset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, featureset: Featureset) -> bool:
        request = UpdateFeaturesetRequest()
        request.featureset.CopyFrom(featureset.raw_message)
        try:
            self.__stub.UpdateFeatureset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Featureset, bool]:
        request = GetFeaturesetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetFeatureset(request)
            return Featureset(response.featureset, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteFeaturesetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteFeatureset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Featureset], bool]:
        request = ListFeaturesetRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListFeaturesets(request)
            return [Featureset(item, self) for item in response.featuresets.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import CronReport as MDCronReport
from github.com.metaprov.modelaapi.services.cronreport.v1.cronreport_pb2_grpc import CronReportServiceStub
from github.com.metaprov.modelaapi.services.cronreport.v1.cronreport_pb2 import CreateCronReportRequest, \
    UpdateCronReportRequest, \
    DeleteCronReportRequest, GetCronReportRequest, ListCronReportsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class CronReport(Resource):
    def __init__(self, item: MDCronReport = MDCronReport(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class CronReportClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: CronReportServiceStub = stub

    def create(self, cronreport: CronReport) -> bool:
        request = CreateCronReportRequest()
        request.cronreport.CopyFrom(cronreport.raw_message)
        try:
            response = self.__stub.CreateCronReport(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, cronreport: CronReport) -> bool:
        request = UpdateCronReportRequest()
        request.cronreport.CopyFrom(cronreport.raw_message)
        try:
            self.__stub.UpdateCronReport(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[CronReport, bool]:
        request = GetCronReportRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetCronReport(request)
            return CronReport(response.cronreport, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteCronReportRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteCronReport(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[CronReport], bool]:
        request = ListCronReportsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListCronReports(request)
            return [CronReport(item, self) for item in response.cronreports.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import Commit as MDCommit
from github.com.metaprov.modelaapi.services.commit.v1.commit_pb2_grpc import CommitServiceStub
from github.com.metaprov.modelaapi.services.commit.v1.commit_pb2 import CreateCommitRequest, \
    UpdateCommitRequest, \
    DeleteCommitRequest, GetCommitRequest, ListCommitsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Commit(Resource):
    def __init__(self, item: MDCommit = MDCommit(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class CommitClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: CommitServiceStub = stub

    def create(self, commit: Commit) -> bool:
        request = CreateCommitRequest()
        request.commit.CopyFrom(commit.raw_message)
        try:
            response = self.__stub.CreateCommit(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, commit: Commit) -> bool:
        request = UpdateCommitRequest()
        request.commit.CopyFrom(commit.raw_message)
        try:
            self.__stub.UpdateCommit(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Commit, bool]:
        request = GetCommitRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetCommit(request)
            return Commit(response.commit, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteCommitRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteCommit(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Commit], bool]:
        request = ListCommitsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListCommits(request)
            return [Commit(item, self) for item in response.commits.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import asyncio
import time

import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import Study as MDStudy
from github.com.metaprov.modelaapi.services.study.v1.study_pb2_grpc import StudyServiceStub
from github.com.metaprov.modelaapi.services.study.v1.study_pb2 import CreateStudyRequest, \
    UpdateStudyRequest, DeleteStudyRequest, GetStudyRequest, ListStudyRequest, \
    GetStudyProfileRequest, CreateStudyProfileRequest, AbortStudyRequest, PauseStudyRequest, ResumeStudyRequest
from tqdm import tqdm

import modela
from modela.common import *
from modela.training.models import ModelSearch, FeatureEngineeringSearch, BaselineSettings, Ensemble, Training, \
    Interpretability, StudySchedule, NotificationSettings, StudySpec, StudyStatus, DataLocation
from modela.Resource import Resource
from modela.ModelaException import ModelaException, ResourceNotFoundException
from typing import List, Union

from modela.data.Dataset import Dataset
from modela.infra.Lab import Lab
from modela.infra.VirtualBucket import VirtualBucket
from modela.training.Model import Model
from modela.training.common import *


class Study(Resource):
    def __init__(self, item: MDStudy = MDStudy(), client=None, namespace="", name="", version=Resource.DefaultVersion,
                 dataset: Union[str, Dataset] = "",
                 test_dataset: Union[str, Dataset] = "",
                 lab: Union[ObjectReference, Lab, str] = "default-lab",
                 bucket: Union[VirtualBucket, str] = None,
                 objective: Metric = None,
                 search: ModelSearch = None,
                 fe_search: FeatureEngineeringSearch = None,
                 baseline: BaselineSettings = None,
                 ensemble: Ensemble = None,
                 trainer_template: Training = None,
                 interpretability: Interpretability = None,
                 schedule: StudySchedule = None,
                 notification: NotificationSettings = None,
                 garbage_collect: bool = True,
                 keep_best_models: bool = True,
                 fast: bool = False,
                 timeout: int = 600,
                 template: bool = False):
        """

        :param client: The Study client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param dataset: If specified as a string, the SDK will attempt to find a Dataset resource with the given name.
            If specified as a Dataset object, or if one was found with the given name, it will be used in the Study.
        :param test_dataset: If specified, the Dataset will be used as the test dataset and the Dataset specified by
            the `dataset` field will be used solely as the training dataset
        :param lab: The object reference, Lab object, or lab name under the tenant of the resource for which all Study-related
            workloads will be performed under.
        :param bucket: The Bucket object or name of the bucket which will store the Study artifacts
        :param objective: The objective metric relevant to the task type.
        :param search: The search parameters define how many models to sample
        :param fe_search: The feature engineering search parameters of the Study
        :param baseline: The baseline settings for the Study, which if enabled will train an unoptimized model of each
            algorithm type for benchmarking.
        :param ensemble: The ensemble settings for the Study, which if enabled will combine the top estimators of
            the study after the initial model search.
        :param trainer_template: The training template for each model created by the Study.
        :param interpretability: The interpretability configuration for the Study, which specifies what type of
            model explainability plots will be generated from Shap values if they are computed
        :param schedule: The schedule for the study to run chronically
        :param notification: The notification settings, which if enabled will forward events about this resource to a notifier.
        :param garbage_collect: If enabled, models which did not move past the testing stage will be garbage collected by
            the system.
        :param keep_best_models: If enabled, the best models from each algorithm will not be garbage collected.
        :param fast: If enabled, the Study will skip profiling, explaining, and reporting.
        :param timeout: The timeout in seconds after which the Study will fail.
        :param template: If the Study is a template it will not start a search and can only be used as a template for
            other studies.
        """
        self.default_resource = False
        super().__init__(item, client, namespace=namespace, name=name, version=version)
        if not self.default_resource:  # Ignore the rest of the constructor; studies are immutable
            return

        spec = self.spec
        if type(dataset) == Dataset:
            dataset_name = dataset.name
        else:
            dataset_name = dataset
            dataset = client.modela.Dataset(namespace=namespace, name=dataset)
            if dataset.default_resource:
                raise ResourceNotFoundException("Dataset {0} does not exist".format(dataset_name))

        spec.DatasetName = dataset_name
        spec.Task = dataset.spec.Task

        if test_dataset:
            if type(test_dataset) == Dataset:
                test_dataset_name = dataset.name
            else:
                test_dataset_name = dataset
                test_dataset = client.modela.Dataset(namespace=namespace, name=dataset)
                if test_dataset.default_resource:
                    raise ResourceNotFoundException("Dataset {0} does not exist".format(dataset_name))

            spec.TrainingTemplate.Split.TrainDataset = dataset_name
            spec.TrainingTemplate.Split.TestDataset = test_dataset_name
            spec.TrainingTemplate.Split.Method = DataSplitMethod.TestDataset

        if type(lab) == Lab:
            lab = lab.reference
        elif type(lab) == str:
            lab = ObjectReference(Namespace=client.modela.tenant, Name=lab)

        if bucket is not None:
            if type(bucket) == VirtualBucket:
                bucket = bucket.name
            spec.Location = DataLocation(BucketName=bucket)

        if objective is not None:
            spec.Search.Objective = objective

        if search is not None:
            spec.Search = search

        if fe_search is not None:
            spec.FeSearch = fe_search

        if baseline is not None:
            spec.Baseline = baseline

        if ensemble is not None:
            spec.Ensembles = ensemble

        if trainer_template is not None:
            spec.TrainingTemplate = trainer_template

        spec.TrainingTemplate.LabRef = lab

        if interpretability is not None:
            spec.Interpretability = interpretability

        if schedule is not None:
            spec.Schedule = schedule

        if notification is not None:
            spec.Notification = notification

        if garbage_collect:
            spec.Gc.CollectAtStudyEnd = True

        if keep_best_models:
            spec.Gc.KeepOnlyBestModelPerAlgorithm = True

        spec.Fast = fast
        # spec.ActiveDeadlineSeconds = timeout
        spec.Template = template

    @property
    def spec(self) -> StudySpec:
        return StudySpec().copy_from(self._object.spec)

    @property
    def status(self) -> StudyStatus:
        self.sync()
        return StudyStatus().copy_from(self._object.status)

    def default(self):
        self.default_resource = True
        StudySpec().apply_config(self._object.spec)

    def abort(self):
        self.ensure_client_repository()
        self._client.abort(self.namespace, self.name)

    def pause(self):
        self.ensure_client_repository()
        self._client.pause(self.namespace, self.name)

    def resume(self):
        self.ensure_client_repository()
        self._client.resume(self.namespace, self.name)

    def submit_and_visualize(self, replace: bool = False, show_progress_bar=True):
        """
        Submit the resource and call visualize().

        :param replace: Replace the resource if it already exists on the cluster.
        :param show_progress_bar: If enabled, the visualization will render a progress bar indicating the study progress.
        """
        self.submit(replace)
        self.visualize(show_progress_bar)

    def visualize(self, show_progress_bar=True):
        """
        Display a real-time visualization of the Study's progress

        :param show_progress_bar: If enabled, the visualization will render a progress bar indicating the study progress.
        """
        desc = tqdm(total=0, position=0, bar_format='{desc}Time Elapsed: {elapsed}')
        bars = {}

        current_status, objective, cv_top, alg_top = None, self.spec.Search.Objective, 0, ""
        try:
            while True:
                status = self.status
                if current_status != status:
                    current_status = status
                    desc.set_description('Phase: %s | Active Models: %d | Best Algorithm: %s | Best Score: %s' %
                                         (StudyPhaseToProgress[self.phase], current_status.Models,
                                          alg_top if alg_top != "" else "[Waiting]",
                                          str(cv_top) if cv_top != 0 else "[Waiting]"))

                    desc.refresh()

                if current_status.Phase == StudyPhase.Failed:
                    desc.set_description("Phase: Failed | Error Message: {0}".format(current_status.FailureMessage))
                    desc.colour = "red"
                    desc.refresh()
                    time.sleep(5)
                    self.sync()
                    if self.phase != StudyPhase.Failed:
                        continue
                    else:
                        break
                elif current_status.Phase == StudyPhase.Completed:
                    model = self._client.modela.Model(self.namespace, current_status.BestModel)
                    if model.phase not in (ModelPhase.Completed, ModelPhase.Live):
                        continue

                    desc.close()
                    for bar in bars.values():
                        bar.close()

                    if not show_progress_bar:
                        desc.set_description_str("Study completed!")
                        desc.refresh()

                    print("\n\n" + model.details)
                    if self.spec.Search.Trainers <= 2:
                        print("Want quicker training speeds? Get more parallel trainers at https://modela.ai")

                    time.sleep(0.1)
                    return

                time.sleep(0.1)
                if not show_progress_bar:
                    time.sleep(0.9)
                    for model in self.models:
                        if len(model.status.Cv) > 0:
                            if model.get_cv_metric(objective) > cv_top:
                                cv_top, alg_top = model.get_cv_metric(objective), model.spec.Estimator.AlgorithmName

                    continue

                models = self.models
                for model in dict(bars).keys():
                    bar = bars[model]
                    mod = [x for x in models if x.name == model]
                    if len(mod) == 0:  # Model has been garbage collected, clean up the bar
                        if bar.pos == -(len(bars) + 1):  # If the bar is in the last position, then clean it up normally
                            bar.close()
                        else:  # Otherwise, we need to shift all the bars down by one
                            _bars = [[mod, _bar] for mod, _bar in bars.items() if _bar.pos <= bar.pos]
                            if len(_bars) <= 1:  # Not supposed to happen; close the bar and move on
                                bar.close()
                            else:
                                _bars.sort(key=lambda bp: bp[1].pos, reverse=True)
                                for idx in range(len(_bars) - 1):  # Copy bar data down the list and fix indexes
                                    bar_forward, bar_current = _bars[idx + 1][1], _bars[idx][1]
                                    bar_current.desc = bar_forward.desc
                                    bar_current.n = bar_forward.n
                                    bar_current.postfix = bar_forward.postfix
                                    bar_current.refresh()
                                    if idx > 0:
                                        bars[_bars[idx][0]] = _bars[idx - 1][1]

                                bars[_bars[-1][0]] = _bars[-2][1]
                                _bars[-1][1].close()

                        del bars[model]

                for model in models:  # For each model, create a bar/update the bar of it already exists
                    if not (model.name in bars):
                        bars[model.name] = tqdm(total=100, position=len(bars) + 1,
                                                bar_format='{desc} {percentage:3.0f}%|{bar}| {postfix}',
                                                desc=model.name + " [Pending]", ncols=121, initial=0, leave=False)
                    else:
                        status = model.status  # Avoid multiple expensive calls to load configs
                        if bars[model.name].n != status.Progress:
                            bars[model.name].n = status.Progress
                            bars[model.name].refresh()

                        dsc = "{0: <50} [{1}]:".format(model.name + "/" + model.spec.Estimator.AlgorithmName,
                                                       status.Phase.name).ljust(62)
                        postfix = "CV = [Waiting]"
                        if len(status.Cv) > 0:
                            postfix = "CV = {0}".format(model.get_cv_metric(objective))
                            if model.get_cv_metric(objective) > cv_top:
                                cv_top, alg_top = model.get_cv_metric(objective), model.spec.Estimator.AlgorithmName

                        if len(status.Test) > 0:
                            postfix = "CV = {0}, Test = {1}".format(model.get_cv_metric(objective),
                                                                    model.get_test_metric(objective))

                        class unappendable_str:  # Workaround for tqdm appending a comma before the postfix >:(
                            def __init__(self, s):
                                self.s = s

                            def __str__(self):
                                return self.s

                        if bars[model.name].desc != dsc:
                            bars[model.name].set_description_str(dsc)

                        if str(bars[model.name].postfix) != postfix:
                            bars[model.name].postfix = unappendable_str(postfix)
                            bars[model.name].refresh()

        except KeyboardInterrupt:
            pass

    async def wait_until_phase(self, phase: StudyPhase):
        """ Returns a coroutine which blocks until the specified phase is reached, or the Study fails """
        while self.status.Phase not in (phase, StudyPhase.Failed):
            await asyncio.sleep(1/5)


    @property
    def models(self) -> List[Model]:
        """ Get the list of currently active models produced by the Study """
        self.ensure_client_repository()
        return self._client.modela.Models.list(self.namespace, {'study': self.name})

    @property
    def best_model(self) -> Model:
        """ Get the best model selected by the model search process """
        self.ensure_client_repository()
        return self._client.modela.Models.get(self.namespace, self._object.status.bestModel)

    @property
    def phase(self) -> StudyPhase:
        return self.status.Phase


class StudyClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: StudyServiceStub = stub

    def create(self, study: Study) -> bool:
        request = CreateStudyRequest()
        request.study.CopyFrom(study.raw_message)
        try:
            response = self.__stub.CreateStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, study: Study) -> bool:
        request = UpdateStudyRequest()
        request.study.CopyFrom(study.raw_message)
        try:
            self.__stub.UpdateStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Study, bool]:
        request = GetStudyRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetStudy(request)
            return Study(response.study, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteStudyRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str, labels: dict = None) -> Union[List[Study], bool]:
        request = ListStudyRequest()
        request.namespace = namespace
        if labels is not None:
            request.labels.update(labels)

        try:
            response = self.__stub.ListStudies(request)
            return [Study(item, self) for item in response.studies.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def abort(self, namespace: str, name: str) -> bool:
        request = AbortStudyRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.AbortStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def pause(self, namespace: str, name: str) -> bool:
        request = PauseStudyRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.PauseStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def resume(self, namespace: str, name: str) -> bool:
        request = ResumeStudyRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.ResumeStudy(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get_profile(self, namespace: str, name: str, models: int) -> bool:
        request = GetStudyProfileRequest()
        request.namespace = namespace
        request.name = name
        request.models = models
        try:
            response = self.__stub.GetStudyProfile(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def create_profile(self, namespace: str, name: str) -> bool:
        request = CreateStudyProfileRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.CreateStudyProfile(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False



import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import ModelaSystem as MDModelaSystem
from github.com.metaprov.modelaapi.services.modelasystem.v1.modelasystem_pb2_grpc import ModelaSystemServiceStub
from github.com.metaprov.modelaapi.services.modelasystem.v1.modelasystem_pb2 import CreateModelaSystemRequest, \
    UpdateModelaSystemRequest, \
    DeleteModelaSystemRequest, GetModelaSystemRequest, ListModelaSystemsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class ModelaSystem(Resource):
    def __init__(self, item: MDModelaSystem = MDModelaSystem(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class ModelaSystemClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ModelaSystemServiceStub = stub

    def create(self, modelasystem: ModelaSystem) -> bool:
        request = CreateModelaSystemRequest()
        request.modelasystem.CopyFrom(modelasystem.raw_message)
        try:
            response = self.__stub.CreateModelaSystem(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, modelasystem: ModelaSystem) -> bool:
        request = UpdateModelaSystemRequest()
        request.modelasystem.CopyFrom(modelasystem.raw_message)
        try:
            self.__stub.UpdateModelaSystem(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[ModelaSystem, bool]:
        request = GetModelaSystemRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModelaSystem(request)
            return ModelaSystem(response.modelasystem, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteModelaSystemRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteModelaSystem(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[ModelaSystem], bool]:
        request = ListModelaSystemsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListModelaSystems(request)
            return [ModelaSystem(item, self) for item in response.modelasystems.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import asyncio
import io
import json
import os.path
import random
import time

import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Dataset as MDDataset
from github.com.metaprov.modelaapi.services.dataset.v1.dataset_pb2_grpc import DatasetServiceStub
from github.com.metaprov.modelaapi.services.dataset.v1.dataset_pb2 import CreateDatasetRequest, \
    UpdateDatasetRequest, \
    DeleteDatasetRequest, GetDatasetRequest, ListDatasetsRequest, GetDatasetProfileRequest
from tabulate import tabulate
from tqdm import tqdm, trange

from modela.Resource import Resource
from modela.ModelaException import ModelaException, ResourceNotFoundException
from typing import List, Union
import pandas

from modela.common import ObjectReference
from modela.data.common import *
from modela.data.DataSource import DataSource
from modela.data.models import SampleSettings, DatasetSpec, DatasetStatus, DatasetProfile
from modela.infra.models import Workload, NotificationSettings
from modela.infra.Lab import Lab
from modela.infra.VirtualBucket import VirtualBucket
from modela.training.Report import Report
from modela.training.common import TaskType
from modela.util import convert_size


class Dataset(Resource):
    def __init__(self, item: MDDataset = MDDataset(), client=None, namespace="", name="",
                 version=Resource.DefaultVersion,
                 lab: Union[ObjectReference, Lab, str] = "default-lab",
                 gen_datasource: bool = False,
                 target_column: str = None,
                 datasource: Union[DataSource, str] = "",
                 bucket: Union[VirtualBucket, str] = "default-minio-bucket",
                 dataframe: pandas.DataFrame = None,
                 data_file: str = None,
                 data_bytes: bytes = None,
                 workload: Workload = Workload("general-large"),
                 fast: bool = False,
                 sample: SampleSettings = SampleSettings(),
                 task_type: TaskType = None,
                 notification: NotificationSettings = None):
        """
        :param client: The Dataset client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param lab: The object reference, Lab object, or lab name under the tenant of the resource for which all
            Dataset-related workloads will be performed under.
        :param gen_datasource: If true, a Datasource resource will be created from the uploaded dataset and applied to
            the Dataset resource.
        :param target_column: If gen_datasource is enabled, then the target column of the data source must be specified.
        :param datasource: If specified as a string, the SDK will attempt to find a Data Source resource with the given name.
            If specified as a Data Source, or if one was found with the given name, it will be applied to the Dataset.
        :param bucket: The bucket which the raw dataset data will be uploaded to.
        :param dataframe: The Pandas Dataframe will be serialized and uploaded for ingestion with the Dataset resource.
        :param data_file: The file path which will be read and uploaded for ingestion with the Dataset resource.
        :param data_bytes: The raw data as bytes that will be uploaded for  ingestion with the Dataset resource.
        :param workload: The resource requirements which will be allocated for Dataset ingestion.
        :param fast: If enabled, the Dataset will skip validation, profiling, and reporting.
        :param sample: The sample settings of the dataset, which if enabled will select only a portion of the Dataset
          for further processing.
        :param task_type: The target task type in relation to the data being used.
        :param notification: The notification settings, which if enabled will forward events about this resource to a notifier.
        """
        self.default_resource = False
        self._profile = None
        super().__init__(item, client, namespace=namespace, name=name, version=version)
        if not self.default_resource:  # Ignore the rest of the constructor; datasets are immutable
            return

        spec = self.spec
        if type(lab) == Lab:
            lab = lab.reference
        elif type(lab) == str:
            lab = ObjectReference(Namespace=client.modela.tenant, Name=lab)
        spec.LabRef = lab

        if type(bucket) == VirtualBucket:
            bucket = bucket.name

        if gen_datasource:
            file_type = FlatFileType.Csv
        else:
            if type(datasource) == str:  # Fetch the data source in case we need to read the file type
                assert datasource != ""
                datasource = client.modela.DataSource(namespace=namespace, name=datasource)
                assert not datasource.default_resource

            file_type = datasource.spec.FileType

        if data_file is not None:
            with open(data_file, 'r') as f:
                data_bytes = f.read()
            data_file = os.path.basename(data_file)
        elif dataframe is not None:
            writer = io.BytesIO()
            if file_type == FlatFileType.Csv:
                df_encoded = dataframe.to_csv(index=False)
            elif file_type == FlatFileType.Parquet:
                df_encoded = dataframe.to_parquet(index=False)
            elif file_type == FlatFileType.Excel:
                dataframe.to_excel(writer)
                df_encoded = writer.getvalue()
            elif file_type == FlatFileType.Json:
                df_encoded = dataframe.to_json()
            elif file_type == FlatFileType.Feather:
                dataframe.to_feather(writer)
                df_encoded = writer.getvalue()
            else:
                raise TypeError(
                    "Pandas cannot deserialize a dataframe to the file type {0}. "
                    "Consider changing your data source file format.".format(file_type))
            data_bytes = bytes(df_encoded, encoding='utf-8')

        if data_bytes is not None:
            self.spec.Origin = client.modela.FileService.upload_file(data_file or name, data_bytes,
                                                                     client.modela.tenant,
                                                                     namespace, version, bucket, "datasets", name)

        if gen_datasource:
            datasource = client.modela.DataSource(namespace=namespace, name=name + "-source", version=version,
                                                  infer_bytes=data_bytes, target_column=target_column, task_type=task_type)
            datasource.submit(replace=True)

        self.spec.DatasourceName = datasource.name
        self.spec.Resources = workload
        self.spec.Task = task_type or TaskType.BinaryClassification
        if not task_type:
            print("WARNING: No task type was specified. Defaulting to binary classification")
        elif task_type != datasource.spec.Task:
            print("WARNING: Dataset task type does not match the task type of its Data Source (you: %s, it: %s)" %
                  (task_type.name, datasource.spec.Task.name))

        self.spec.Fast = fast
        if sample is not None:
            self.spec.Sample = sample

        if notification is not None:
            self.spec.Notification = notification

    @property
    def spec(self) -> DatasetSpec:
        return DatasetSpec().copy_from(self._object.spec).set_parent(self._object.spec)

    @property
    def status(self) -> DatasetStatus:
        self.sync()
        return DatasetStatus().copy_from(self._object.status)

    def default(self):
        self.default_resource = True
        DatasetSpec().apply_config(self._object.spec)

    @property
    def report(self) -> Report:
        """ Fetch the report associated with the Dataset """
        self.ensure_client_repository()
        if self.status.ReportName == "":
            raise ValueError("Dataset {0} has no report.".format(self.name))

        return self._client.modela.Report(namespace=self.namespace, name=self.status.ReportName)

    @property
    def phase(self) -> DatasetPhase:
        """ The phase specified by the status of the Dataset """
        return self.status.Phase

    @property
    def datasource(self) -> DataSource:
        """ Fetch the DataSource object associated with the Dataset """
        self.ensure_client_repository()
        return self._client.modela.DataSource(self.namespace, self.spec.DatasourceName)

    @property
    def test_prediction(self) -> str:
        """ Generate a default prediction payload for a model that would be created with the dataset. """
        target = self.datasource.target_column.Name
        return json.dumps(
            [{col.Name: self.status.Statistics.column(col.Name).Mean if col.Datatype == DataType.Number else col.Enum[0]
              for col in self.datasource.spec.Schema.Columns if col.Name != target}])

    def submit_and_visualize(self, replace: bool = False, show_progress_bar=True):
        """
        Submit the resource and call visualize().

        :param replace: Replace the resource if it already exists on the cluster.
        :param show_progress_bar: If enabled, the visualization will render a progress bar indicating the study progress.
        """
        self.submit(replace)
        self.visualize(show_progress_bar)

    def visualize(self, show_progress_bar=True):
        """
        Display a real-time visualization of the Dataset's progress

        :param show_progress_bar: If enabled, the visualization will render a progress bar indicating the dataset's progress.
        """
        desc, progress = tqdm(total=0, position=0, bar_format='{desc}  Time Elapsed: {elapsed}'), None
        if show_progress_bar:
            progress = tqdm(total=100, position=1, bar_format='{l_bar}{bar}',
                            desc=self.name, ncols=80, initial=0)

        current_status = self.status
        try:
            while True:
                if current_status == self.status:
                    time.sleep(0.1)
                    continue

                current_status = self.status
                desc.set_description_str('Phase: %s | File Size: %s | Rows: %s | Columns: %s' %
                                         (self.phase.name, convert_size(self.status.Statistics.FileSize),
                                          "[Processing]" if self.status.Statistics.Rows == 0 else self.status.Statistics.Rows,
                                          "[Processing]" if len(self.status.Statistics.Columns) == 0 else len(
                                              self.status.Statistics.Columns)))

                if self.phase in (DatasetPhase.Ready, DatasetPhase.Aborted, DatasetPhase.Failed):
                    if self.phase == DatasetPhase.Ready:
                        print("\n\n" + self.profile)
                    else:
                        if progress:
                            progress.colour = "red"
                            progress.refresh()
                        print("\nDataset was failed or aborted: %s" % current_status.FailureMessage)

                    break

                if progress:
                    progress.n = self.status.Progress
                    progress.last_print_n = self.status.Progress
                    progress.refresh()

        except KeyboardInterrupt:
            pass

        desc.close()
        if progress:
            progress.close()

    @property
    def profile(self) -> str:
        """
        Returns an ASCII-rendered table of the Dataset's profile
        """
        profile = self.get_profile()

        table = []
        for col in profile.Columns:
            table.append([col.Name, col.Type, col.Distinct, col.Missing,
                          '{0:.3g}'.format(col.Mean), '{0:.3g}'.format(col.Stddev), col.P50, col.Min, col.Max,
                          '{0:.3g}'.format(col.CorrToTarget), '{0:.3g}'.format(col.Variance),
                          '{0:.3g}'.format(col.Skewness), '{0:.3g}'.format(col.Kurtosis)])

        table = tabulate(table, tablefmt='pretty', headers=['Column', 'Data Type', 'Distinct', 'Missing', 'Mean',
                                                            'Stddev', 'Median', 'Min', 'Max', 'Corr. To Target',
                                                            'Variance', 'Skewness', 'Kurtosis'])
        return table + "\n"
        # TODO: jupyter support, print all viz plots in jupyter

    @property
    def details(self) -> str:
        """
        Returns an ASCII-rendered table of the Dataset's statistics
        """
        table = []
        for col in self.status.Statistics.Columns:
            table.append([col.Name, col.Datatype.name, col.Distinct, col.Missing,
                          '{0:.3g}'.format(col.Mean), '{0:.3g}'.format(col.Stddev), col.P50, col.Min, col.Max,
                          '{0:.3g}'.format(col.Skewness), '{0:.3g}'.format(col.Kurtosis)])

        table = tabulate(table, tablefmt='pretty', headers=['Column', 'Data Type', 'Distinct', 'Missing', 'Mean',
                                                            'Stddev', 'Median', 'Min', 'Max', 'Skewness', 'Kurtosis'])
        return table + "\n"

    def get_profile(self) -> DatasetProfile:
        if self._profile:
            return self._profile
        if hasattr(self, "_client"):
            self._profile = self._client.profile(self.namespace, self.name)
            return self._profile
        else:
            raise AttributeError("Object has no client repository")

    async def wait_until_phase(self, phase: DatasetPhase):
        """ Returns a coroutine which blocks until the specified phase is reached, or the Dataset fails """
        while self.status.Phase not in (phase, DatasetPhase.Failed):
            await asyncio.sleep(1/5)

    def __repr__(self):
        out = super().__repr__()
        if len(self.status.Statistics.Columns) > 0:
            out += "\n" + self.details

        return out


class DatasetClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DatasetServiceStub = stub

    def create(self, dataset: Dataset) -> bool:
        request = CreateDatasetRequest()
        request.dataset.CopyFrom(dataset.raw_message)
        try:
            response = self.__stub.CreateDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, dataset: Dataset) -> bool:
        request = UpdateDatasetRequest()
        request.dataset.CopyFrom(dataset.raw_message)
        try:
            self.__stub.UpdateDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Dataset:
        request = GetDatasetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataset(request)
            return Dataset(response.dataset, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDatasetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> List[Dataset]:
        request = ListDatasetsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDatasets(request)
            return [Dataset(item, self) for item in response.datasets.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)

    def profile(self, namespace: str, name: str) -> DatasetProfile:
        request = GetDatasetProfileRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDatasetProfile(request)
            return DatasetProfile().copy_from(response.profile)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False



import grpc
from github.com.metaprov.modelaapi.pkg.apis.team.v1alpha1.generated_pb2 import Review as MDReview
from github.com.metaprov.modelaapi.services.review.v1.review_pb2_grpc import ReviewServiceStub
from github.com.metaprov.modelaapi.services.review.v1.review_pb2 import CreateReviewRequest, \
    UpdateReviewRequest, \
    DeleteReviewRequest, GetReviewRequest, ListReviewRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Review(Resource):
    def __init__(self, item: MDReview = MDReview(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class ReviewClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ReviewServiceStub = stub

    def create(self, review: Review) -> bool:
        request = CreateReviewRequest()
        request.review.CopyFrom(review.raw_message)
        try:
            response = self.__stub.CreateReview(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, review: Review) -> bool:
        request = UpdateReviewRequest()
        request.review.CopyFrom(review.raw_message)
        try:
            self.__stub.UpdateReview(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Review, bool]:
        request = GetReviewRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetReview(request)
            return Review(response.review, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteReviewRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteReview(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Review], bool]:
        request = ListReviewRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListReviews(request)
            return [Review(item, self) for item in response.reviews.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import FeaturePipelineRun as MDFeaturePipelineRun
from github.com.metaprov.modelaapi.services.featurepipelinerun.v1.featurepipelinerun_pb2_grpc import FeaturePipelineRunServiceStub
from github.com.metaprov.modelaapi.services.featurepipelinerun.v1.featurepipelinerun_pb2 import CreateFeaturePipelineRunRequest, \
    UpdateFeaturePipelineRunRequest, \
    DeleteFeaturePipelineRunRequest, GetFeaturePipelineRunRequest, ListFeaturePipelineRunRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class FeaturePipelineRun(Resource):
    def __init__(self, item: MDFeaturePipelineRun = MDFeaturePipelineRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class FeaturePipelineRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: FeaturePipelineRunServiceStub = stub

    def create(self, featurepipelinerun: FeaturePipelineRun) -> bool:
        request = CreateFeaturePipelineRunRequest()
        request.featurepipelinerun.CopyFrom(featurepipelinerun.raw_message)
        try:
            response = self.__stub.CreateFeaturePipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, featurepipelinerun: FeaturePipelineRun) -> bool:
        request = UpdateFeaturePipelineRunRequest()
        request.featurepipelinerun.CopyFrom(featurepipelinerun.raw_message)
        try:
            self.__stub.UpdateFeaturePipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[FeaturePipelineRun, bool]:
        request = GetFeaturePipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetFeaturePipelineRun(request)
            return FeaturePipelineRun(response.featurepipelinerun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteFeaturePipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteFeaturePipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[FeaturePipelineRun], bool]:
        request = ListFeaturePipelineRunRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListFeaturePipelineRuns(request)
            return [FeaturePipelineRun(item, self) for item in response.featurepipelineruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False


